"""
Mathematical and physical constants
"""

from caeroc._skaero import units as u

# Earth's mean equatorial radius from IAU Working Group on Cartographic
# Coordinates and Rotational Elements: 2015
R_earth = 6.3781366e6 * u.meters

# Standard Earth's gravity acceleration value from "The International System of
# Units by Barry N.Taylor and Ambler Thompson"
g0 = 9.80665 * u.meters / u.seconds**2



"""
Decorators for multiple purposes.

"""

from functools import update_wrapper, wraps


def implicit(f):
    """Returns implicit function from a single-variable function.

    The form of the implicit function is

    f(t) -> F(t; x) = f(t) - x

    so the equation F(t; x) = 0 gives x = f(t).

    """

    @wraps(f)
    def _F(t, y):
        return f(t) - y

    return _F



import sys

from caeroc.gui import runtime

try:
    from PySide2.QtWidgets import QApplication as _QtApp
except ImportError:
    from PyQt5.QtWidgets import QApplication as _QtApp


class CalcApp:
    def __init__(self):
        try:
            self.app = _QtApp(sys.argv)
        except RuntimeError:
            self.app = _QtApp.instance()  # if QApplication already exists
        self.dialog = runtime.CalcDialog()

    def run(self):
        self.dialog.show()
        sys.exit(self.app.exec_())


if __name__ == "__main__":
    CalcApp().run()



"""
Gas dynamics calculations.

The following notation is used throughout the different modules:

* `X_0` property X at the stagnation point.
* `X_star` property X at the critical point.
* `X_1` upstream property X (when applicable).
* `X_2` downstream property X (when applicable).
* `X_X0` ratio between `X` and `X_0`.
* `X2_X1` ratio between `X_2` and `X_1`.

"""



import unittest

import numpy as np

import caeroc.formulae as fml


class TestIsentropic(unittest.TestCase):

    def setUp(self):
        self.mode = fml.isentropic.Isentropic(1.5)

    def test_basic(self):
        """Test all class methods in Isentropic class"""
        isen = self.mode
        isen.Mt(M=0.5)
        isen.rho_rho0(M=0.5, store=True)
        isen.p_pt(M=0.5, store=False)
        isen.calculate(A_Astar=2.0)

    def test_calculate(self):
        isen = self.mode
        isen.calculate(M=2)
        self.assertAlmostEqual(isen.data["p_p0"][0], 0.125)
        self.assertAlmostEqual(isen.data["rho_rho0"][0], 0.25)
        self.assertAlmostEqual(isen.data["p_pt"][0], 0.24414062)
        self.assertAlmostEqual(isen.data["rho_rhot"][0], 0.390625)
        self.assertAlmostEqual(isen.data["A_Astar"][0], 1.61908616)

    def test_calculate_array_like(self):
        x = np.linspace(0, 5, 500, endpoint=True)
        self.mode.calculate(M=x)


if __name__ == "main":
    unittest.main()



"""
caeroc
========

Provides

  1. An object-oriented calculator utility and plotting.
  2. GUI interface.

The docstring examples assume that `fluiddyn` has been imported as below::

  >>> import caeroc

Use the built-in ``help`` function to view a function's docstring::

  >>> help(caeroc.gui)
  ... # doctest: +SKIP

Available subpackages
---------------------
formulae
    Governing integral equations for compressible flows organised as various
    classes.
gui
    A PyQt/PySide based interface to have an user-friendly view.
profiles
    Toolkit for making specific calculations for a body.

"""

__version__ = "0.0.2b2"


def launch():
    from . import gui

    gui.CalcApp().run()



import matplotlib.pyplot as plt
import numpy as np


def zeta_coord(radius, centre_chi=0.0, centre_eta=0.0):
    """Transform a cicle into zeta coordinates."""
    angles = np.linspace(0, 2.0 * pl.pi, 100)
    chi = radius * np.cos(angles) + centre_chi
    eta = radius * np.sin(angles) + centre_eta
    zeta = chi + 1j * eta
    return zeta


def z_coord(zeta=None, radius=1.2, centre_chi=-0.02, centre_eta=0.01, plot=False):
    """Joukowsky transformation as a complex number."""
    if zeta is None:
        zeta = zeta_coord(radius, centre_chi, centre_eta)

    z = zeta + 1.0 / zeta
    x = z.real
    y = z.imag
    if plot:
        plt.plot(x, y)
        plt.show()

    return z



"""
Compressible Aerodynamics calculator
"""

import math as Math

import matplotlib.pyplot as plt
import numpy as np


class Body:
    def __init__(self, N, length, bodytype=None):
        self.N = N
        self.L = length
        if bodytype is "airfoil":
            self.X, self.Y, self.angle = self.airfoil()
        else:
            self.X, self.Y, self.angle = self.circle()
        plt.plot(self.X, self.Y)
        plt.axis([-length, length, -length, length])
        self._non_dimensionalize(length)

    def nozzle(self):
        """Converging/Diverging Nozzle"""
        pass

    def lavalnozzle(self):
        """Conv. and Div. Nozzle"""
        pass

    def airfoil(self):
        """Biconvex airfoil"""
        L = self.L
        N = self.N
        R = 210
        theta_arc = np.arcsin(L / 2 / R) * 2
        pi = np.pi
        theta_up = np.linspace((pi + theta_arc) / 2, +(pi - theta_arc) / 2, N)
        # theta_up = theta[::-1]
        theta_down = np.linspace((3 * pi - theta_arc) / 2, +(3 * pi + theta_arc) / 2, N)

        X_up = R * np.cos(theta_up)
        Y_up = R * np.sin(theta_up)
        X_down = R * np.cos(theta_down)
        Y_down = R * np.sin(theta_down)

        shift_r = X_up[0]
        X_up -= shift_r
        X_down -= shift_r
        shift_up = Y_up[0]
        shift_down = Y_down[0]
        Y_up = Y_up - shift_up
        Y_down = Y_down - shift_down

        X = np.concatenate((X_up, X_down))
        Y = np.concatenate((Y_up, Y_down))
        slope_up = np.gradient(Y_up, 1) / np.gradient(X_up, 1)
        slope_down = np.gradient(Y_down, 1) / np.gradient(X_down, 1)
        angle = np.arctan(np.concatenate((slope_up, slope_down)))
        return X, Y, angle

    def circle(self):
        theta = np.linspace(0, 2 * np.pi, self.N)
        R = self.L / 2
        X = R * np.cos(theta)
        Y = R * np.sin(theta)
        angle = theta - np.pi / 2
        return X, Y, angle

    def duct(self):
        """Constant area duct/pipe for Fanno & Rayleigh flows"""
        pass

    def _non_dimensionalize(self, scale):
        """Non dimensionalize all geometric parameters with the scale"""
        self.L = self.L / scale
        self.X = self.X / scale
        self.Y = self.Y / scale


class Isentropic:
    def p_p0(self, M, gamma=1.4):
        return self.t_t0(M, gamma) ** (-gamma / (gamma - 1))

    def t_t0(self, M, gamma=1.4):
        return 1 + (gamma - 1) / 2 * M**2


class Expansion(Isentropic):
    """Isentropic Expansion fan flow relations"""

    def M2(self, M1, theta, gamma=1.4):
        if M1 < 0:
            raise ValueError("Subsonic flow.")
        if theta < 0 or theta > np.pi / 2:
            raise ValueError("Incorrect deflection angle. Cannot calculate!")
        pm1 = self.pm(M1, gamma)
        pm2 = pm1 + theta

        mnew = 2.0
        m = 0.0
        while np.abs(mnew - m) > 0.00001:
            m = mnew
            fm = self.pm(m, gamma) - pm2  # *np.pi/180.
            fdm = np.sqrt(m**2 - 1.0) / (1.0 + 0.5 * (gamma - 1.0) * m**2) / m
            mnew = m - fm / fdm

        M2 = m
        return M2

    def pm(self, M1, gamma=1.4):
        if M1 > 1.0:
            g = gamma
            m = M1
            n = np.sqrt((g + 1.0) / (g - 1.0)) * np.arctan(
                np.sqrt((g - 1.0) / (g + 1.0) * (m * m - 1.0))
            )
            n = n - np.arctan(np.sqrt(m * m - 1.0))
            # n = n * 180./np.pi
        elif M1 == 1:
            n = 0.0
        else:
            n = None

        numax = (np.sqrt((gamma + 1.0) / (gamma - 1.0)) - 1) * 90.0
        if n <= 0.0 or n >= numax:
            raise ValueError("Prandtl-Meyer angle out of bounds")
        return n


class ObliqueShock:
    def beta(self, m1, d, g=1.4, i=0):
        p = -(m1 * m1 + 2.0) / m1 / m1 - g * Math.sin(d) * Math.sin(d)
        q = (2.0 * m1 * m1 + 1.0) / Math.pow(m1, 4.0) + (
            (g + 1.0) * (g + 1.0) / 4.0 + (g - 1.0) / m1 / m1
        ) * Math.sin(d) * Math.sin(d)
        r = -Math.cos(d) * Math.cos(d) / Math.pow(m1, 4.0)

        a = (3.0 * q - p * p) / 3.0
        b = (2.0 * p * p * p - 9.0 * p * q + 27.0 * r) / 27.0

        test = b * b / 4.0 + a * a * a / 27.0

        if test > 0.0:
            return -1.0
        elif test == 0.0:
            x1 = Math.sqrt(-a / 3.0)
            x2 = x1
            x3 = 2.0 * x1
            if b > 0.0:
                x1 *= -1.0
                x2 *= -1.0
                x3 *= -1.0

        if test < 0.0:
            phi = Math.acos(Math.sqrt(-27.0 * b * b / 4.0 / a / a / a))
            x1 = 2.0 * Math.sqrt(-a / 3.0) * Math.cos(phi / 3.0)
            x2 = 2.0 * Math.sqrt(-a / 3.0) * Math.cos(phi / 3.0 + np.pi * 2.0 / 3.0)
            x3 = 2.0 * Math.sqrt(-a / 3.0) * Math.cos(phi / 3.0 + np.pi * 4.0 / 3.0)
            if b > 0.0:
                x1 *= -1.0
                x2 *= -1.0
                x3 *= -1.0

        s1 = x1 - p / 3.0
        s2 = x2 - p / 3.0
        s3 = x3 - p / 3.0

        if s1 < s2 and s1 < s3:
            t1 = s2
            t2 = s3
        elif s2 < s1 and s2 < s3:
            t1 = s1
            t2 = s3
        else:
            t1 = s1
            t2 = s2

        b1 = Math.asin(Math.sqrt(t1))
        b2 = Math.asin(Math.sqrt(t2))

        betas = b1
        betaw = b2
        if b2 > b1:
            betas = b2
            betaw = b1

        if i == 0:
            return betaw
        if i == 1:
            return betas

    def M1nM2n(self, M1, beta, gamma=1.4):
        M1n = M1 * np.sin(beta)
        M2n = np.sqrt(
            (1.0 + 0.5 * (gamma - 1.0) * M1n**2)
            / (gamma * M1n**2 - 0.5 * (gamma - 1.0))
        )
        return M1n, M2n

    def M2(self, M1, theta, gamma=1.4):
        beta = self.beta(M1, theta)
        M1n, M2n = self.M1nM2n(M1, beta, gamma)
        M2 = M2n / np.sin(beta - theta)
        return M2

    def p2_p1(self, M1, theta, gamma=1.4):
        beta = self.beta(M1, theta)
        M1n = M1 * np.sin(beta)
        p2_p1 = 1.0 + 2.0 * gamma / (gamma + 1.0) * (M1n**2 - 1.0)
        return p2_p1


class AirfoilChara:
    def __init__(self, N, Mach, AngleOfAttack):
        self.cp_lin = np.empty(N)
        self.cp_shock = np.empty(N)
        self.M1 = Mach
        self.AoA = AngleOfAttack * np.pi / 180

        self.obs = ObliqueShock()
        self.exp = Expansion()

    def calc_cp_lin(self, angle):
        N = angle.size / 2
        angle = angle - self.AoA
        self.cp_lin[:N] = 2 * angle[:N] / np.sqrt(self.M1**2 - 1)
        self.cp_lin[N:] = 2 * -angle[N:] / np.sqrt(self.M1**2 - 1)

    def calc_cp_shock(self, angle, debug=False):
        gamma = 1.4
        N = angle.size

        M1 = self.M1
        M2 = np.zeros(N)
        p2_p1 = np.zeros(N)

        # Oblique shock
        angle = angle - self.AoA
        M2[0] = self.obs.M2(M1, angle[0])
        M2[N / 2] = self.obs.M2(M1, -angle[N / 2])
        p2_p1[0] = self.obs.p2_p1(M1, angle[0])
        p2_p1[N / 2] = self.obs.p2_p1(M1, -angle[N / 2])

        # Expanded flow
        for i in range(0, N / 2 - 1):
            j = i + 1
            k = N / 2 + i
            L = N / 2 + i + 1
            if debug:
                print("M2=", M2[i], M2[k], "   p2/p1=", p2_p1[i], p2_p1[k])

            dangle = angle[i] - angle[j]
            M2[j] = self.exp.M2(M2[i], dangle)
            dangle = -(angle[k] - angle[l])
            M2[L] = self.exp.M2(M2[k], dangle)

            p2_p1[j] = self.exp.p_p0(M2[j]) / self.exp.p_p0(M2[i]) * p2_p1[i]
            p2_p1[L] = self.exp.p_p0(M2[L]) / self.exp.p_p0(M2[k]) * p2_p1[k]

        self.cp_shock = 2 * (p2_p1 - 1) / (gamma * M1**2)
        if debug:
            for i in range(0, N / 2 - 1):
                k = N / 2 + i
                print("Cp=", self.cp_shock[i], self.cp_shock[k])


if __name__ == "__main__":
    N = 100
    length = 70.0
    airfoil = Body(N, length, "airfoil")
    chara = AirfoilChara(2 * N, Mach=2, AngleOfAttack=1.5)

    # plt.subplot(111)
    plt.plot(airfoil.X, airfoil.Y)
    plt.axis([0, 1, -0.3, 0.3])
    plt.show()

    fig, (ax1, ax2) = plt.subplots(1, 2, sharex=True)

    chara.calc_cp_lin(airfoil.angle)
    ax1.set_title("$C_p$ : Linear theory")
    ax1.grid(True)
    ax1.plot(airfoil.X[:N], chara.cp_lin[:N], "r.-")
    ax1.plot(
        airfoil.X[N:],
        chara.cp_lin[N:],
        ".-",
    )

    chara.calc_cp_shock(airfoil.angle, debug=True)
    ax2.set_title("$C_p$ : Nonlinear theory")
    ax2.grid(True)
    ax2.plot(airfoil.X[0:N], chara.cp_shock[0:N], "r.-")
    ax2.plot(airfoil.X[N : 2 * N], chara.cp_shock[N : 2 * N], ".-")
    fig.show()



import numpy as np

try:
    from .._skaero.gasdynamics import isentropic as _isentropic
except ImportError:
    from skaero.gasdynamics import isentropic as _isentropic


from ..logger import logger
from ..util.decorators import storeresult
from .base import FormulaeBase


def safe_div(x1, x2):
    """Quietly divide x1 by x2, while ignoring RuntimeWarnings."""
    old_err_state = np.seterr(divide="ignore")
    out = np.divide(x1, x2)
    np.seterr(**old_err_state)
    return out


class Isentropic(FormulaeBase, _isentropic.IsentropicFlow):
    """Isentropic flow relations for quasi 1D flows"""

    def __init__(self, gamma=1.4):
        self.keys = [
            "M",
            "p_p0",
            "rho_rho0",
            "T_T0",
            "Mt",
            "p_pt",
            "rho_rhot",
            "T_Tt",
            "A_Astar",
        ]
        super().__init__(gamma=gamma)

    @storeresult
    def p_p0(self, M, *args, **kwargs):
        """
        Parameters
        ----------
        M : array_like
            Mach number.
        store : bool
            If set as true stores the result in self.data. Optional.

        Notes
        -------
        .. math::
           \dfrac{p}{p_0} = \dfrac{T}{T_0} ^ (\gamma / (\gamma - 1))

        """
        return super().p_p0(M)

    @storeresult
    def rho_rho0(self, M, *args, **kwargs):
        """
        Parameters
        ----------
        M : array_like
            Mach number.
        store : bool
            If set as true stores the result in self.data. Optional.

        Notes
        -------
        ..  math::
            \dfrac{\rho}{\rho_0} = \dfrac{T}{T_0} ^ (1 / (\gamma - 1))

        """
        return super().rho_rho0(M)

    @storeresult
    def T_T0(self, M, *args, **kwargs):
        """
        Parameters
        ----------
        M : array_like
            Mach number.
        store : bool
            If set as true stores the result in self.data. Optional.

        Notes
        -------
        ..  math::
            \dfrac{T}{T_0} = (1+(\gamma-1)/2 * M^2) ^ {-1}

        """
        return super().T_T0(M)

    @storeresult
    def Mt(self, M, *args, **kwargs):
        """
        Parameters
        ----------
        M : array_like
            Mach number.
        store : bool
            If set as true stores the result in self.data. Optional.

        Notes
        -------
        ..  math::
            M^* = ((\gamma + 1)/(\gamma - 1) * (1./M^2/(\gamma -1) + 0.5))^(-0.5)

        """
        gp1 = self.gamma + 1
        gm1 = self.gamma - 1

        Mt = 1.0 / np.sqrt(gp1 / gm1 * (safe_div(1, M**2) / gm1 + 0.5))
        return Mt

    @storeresult
    def p_pt(self, M, *args, **kwargs):
        """
        Parameters
        ----------
        M : array_like
            Mach number.
        store : bool
            If set as true stores the result in self.data. Optional.

        """
        g = self.gamma
        return super().p_p0(M) * np.power((g / 2.0 + 0.5), g / (g - 1.0))

    @storeresult
    def rho_rhot(self, M, *args, **kwargs):
        """
        Parameters
        ----------
        M : array_like
            Mach number.
        store : bool
            If set as true stores the result in self.data. Optional.

        """
        g = self.gamma
        return super().rho_rho0(M) * np.power((g / 2.0 + 0.5), 1.0 / (g - 1.0))

    @storeresult
    def T_Tt(self, M, *args, **kwargs):
        """
        Parameters
        ----------
        M : array_like
            Mach number.
        store : bool
            If set as true stores the result in self.data. Optional.

        """
        g = self.gamma
        return super().T_T0(M) * (g / 2.0 + 0.5)

    @storeresult
    def A_Astar(self, M, *args, **kwargs):
        """
        Parameters
        ----------
        M : array_like
            Mach number.
        store : bool
            If set as true stores the result in self.data. Optional.

        Notes
        ------
        .. math::
           \dfrac{A}{A^*} = \dfrac{1}{M}
                            \sqrt{\dfrac{2}{(gamma+1} / \dfrac{T}{T_0}(M)
                                     } ^ {(gamma+1)/(gamma-1)}

        """
        return super().A_Astar(M)

    @storeresult
    def M(self, p_p0=None, rho_rho0=None, T_T0=None, A_Astar=None, *args, **kwargs):
        """
        Computes Mach number when one of the arguments are specified

        """
        g = self.gamma
        if p_p0 is not None:
            M = np.sqrt(2.0 * ((1.0 / np.power(p_p0, (g - 1.0) / g)) - 1.0) / (g - 1.0))
        elif rho_rho0 is not None:
            M = np.sqrt(2.0 * ((1.0 / np.power(rho_rho0, (g - 1.0))) - 1.0) / (g - 1.0))
        elif T_T0 is not None:
            M = np.sqrt(2.0 * ((1.0 / T_T0) - 1.0) / (g - 1.0))
        elif A_Astar is not None:
            Msub, Msup = _isentropic.mach_from_area_ratio(A_Astar)
            M = np.array([Msub, Msup])
        elif "M" in kwargs.keys():
            return kwargs["M"]
        else:
            logger.error("Insufficient data to calculate Mach number")

        return M

    def calculate(self, M=None, p_p0=None, rho_rho0=None, T_T0=None, A_Astar=None):
        """
        Wrapper function to calculate all possible data and store
        using keywords and values in the dictionary `data`.

        Parameters
        ----------
        M, p_p0, rho_rho0, t_t0, A_At : array_like
            Input parameters to calculate, optional but specify one.

        """
        if M is not None:
            mach = M
        else:
            kwargs = {
                "p_p0": p_p0,
                "rho_rho0": rho_rho0,
                "T_T0": T_T0,
                "A_Astar": A_Astar,
                "store": False,
            }
            mach = self.M(**kwargs)

        for key in self.keys:
            property_func = getattr(self, key)
            property_func(M=mach, store=True)

        return self.data


class Expansion(FormulaeBase, _isentropic.PrandtlMeyerExpansion):
    """Isentropic expansion fan flow relations"""

    def __init__(self, gamma=1.4):
        self.keys = [
            "M_1",
            "M_2",
            "nu_1",
            "nu_2",
            "theta",
            "p2_p1",
            "rho2_rho1",
            "T2_T1",
        ]
        self.gamma = gamma

    def calculate(self, theta_deg=None, theta_rad=None, M_1=None, nu_1=None):
        """
        Calculate all possible data and store
        using keywords and values in the dictionary `data`.

        Parameters
        ----------
        theta_deg or theta_rad : float
            Turn angle or deflection angle, optional but specify one.

        M_1 or nu_1 : float
            Mach number or Prandtl-Meyer angle (in radians) of inflow

        """
        if theta_deg:
            self.theta = np.radians(theta_deg)
        elif theta_rad:
            self.theta = theta_rad
        else:
            logger.error("Insufficient data: Turn angle must be specified.")

        if M_1 is None:
            if nu_1 is not None:
                M_1 = _isentropic.mach_from_nu(nu=nu_1, gamma=self.gamma)
            else:
                logger.error("Insufficient data: M_1 or nu_1" + "must be specified.")

        self.M_1 = M_1
        super().__init__(M_1=self.M_1, theta=self.theta)
        for key in self.keys:
            self.store(key, getattr(self, key))

        return self.data



import numpy as np

try:
    from .._skaero.gasdynamics.shocks import _ShockClass
except ImportError:
    from skaero.gasdynamics.shocks import _ShockClass

from ..logger import logger
from ..util.decorators import storeresult
from .base import FormulaeBase


class NormalShock(FormulaeBase, _ShockClass):
    """Normal shock relations"""

    def __init__(self, **kwargs):
        if "gamma" in kwargs:
            self.gamma = kwargs["gamma"]
        else:
            self.gamma = kwargs["gamma"] = 1.4

        if "M_1" not in kwargs:
            self.M_1 = kwargs["M_1"] = self.get_M_1(**kwargs)

        if "beta" not in kwargs:
            # Strong shock solution
            kwargs["beta"] = np.pi / 2

        self.keys = [
            "M_1",
            "M_2",
            "p2_p1",
            "rho2_rho1",
            "T2_T1",
            "p02_p01",
            "rho02_rho01",
            "T02_T01",
        ]

        super().__init__(**kwargs)

    def get_M_1(
        self,
        M_2=None,
        p2_p1=None,
        rho2_rho1=None,
        T2_T1=None,
        p02_p01=None,
        p2_p01=None,
        **kwargs,
    ):
        """
        Computes Mach number when one of the arguments are specified

        """
        try:
            g = self.gamma
        except KeyError:
            g = kwargs["gamma"]

        if p2_p1 is not None:
            M_1 = np.sqrt((p2_p1 - 1) * (g + 1.0) / 2.0 / g + 1.0)
        elif rho2_rho1 is not None:
            M_1 = np.sqrt(2.0 * rho2_rho1 / (g + 1.0 - rho2_rho1 * (g - 1.0)))
        elif T2_T1 is not None:
            a = 2.0 * g * (g - 1.0)
            b = 4.0 * g - (g - 1.0) * (g - 1.0) - T2_T1 * (g + 1.0) * (g + 1.0)
            c = -2.0 * (g - 1.0)
            M_1, M_11 = np.roots([a, b, c])
        elif p02_p01 is not None:
            raise NotImplementedError
        elif p2_p01 is not None:
            raise NotImplementedError
        elif "M" in kwargs.keys():
            return kwargs["M"]
        else:
            logger.error("Insufficient data to calculate Mach number")

        return M_1

    def calculate(
        self,
        M_1=None,
        M_2=None,
        p2_p1=None,
        rho2_rho1=None,
        T2_T1=None,
        p02_p01=None,
        rho02_rho01=None,
        T02_T01=None,
    ):
        """
        Calculate all possible data and store
        using keywords and values in the dictionary `data`.

        Parameters
        ----------
        theta_deg or theta_rad : float
            Turn angle or deflection angle, optional but specify one.

        M_1: float
            Mach number of inflow

        """
        if M_1 is not None:
            self.M_1 = M_1
        elif any([v is not None for v in (p2_p1, rho2_rho1, T2_T1)]):
            self.M_1 = self.get_M_1(p2_p1, rho2_rho1, T2_T1)
        else:
            logger.error(
                "Insufficient data: M_1 or p2_p1 or rho2_rho1 "
                "or T2_T1 must be specified."
            )

        super().__init__(M_1=self.M_1, beta=np.pi / 2, gamma=self.gamma)
        for key in self.keys:
            self.store(key, getattr(self, key))

        return self.data


class ObliqueShock(NormalShock):
    def __init__(self, gamma=1.4):
        self.keys = [
            "M_1",
            "M_2",
            "M_1n",
            "M_2n",
            "theta",
            "p2_p1",
            "rho2_rho1",
            "T2_T1",
        ]
        self.gamma = gamma

    def beta(self, m1, d, g=1.4, i=0):
        p = -(m1 * m1 + 2.0) / m1 / m1 - g * np.sin(d) * np.sin(d)
        q = (2.0 * m1 * m1 + 1.0) / np.pow(m1, 4.0) + (
            (g + 1.0) * (g + 1.0) / 4.0 + (g - 1.0) / m1 / m1
        ) * np.sin(d) * np.sin(d)
        r = -np.cos(d) * np.cos(d) / np.pow(m1, 4.0)

        a = (3.0 * q - p * p) / 3.0
        b = (2.0 * p * p * p - 9.0 * p * q + 27.0 * r) / 27.0

        test = b * b / 4.0 + a * a * a / 27.0

        if test > 0.0:
            return -1.0
        elif test == 0.0:
            x1 = np.sqrt(-a / 3.0)
            x2 = x1
            x3 = 2.0 * x1
            if b > 0.0:
                x1 *= -1.0
                x2 *= -1.0
                x3 *= -1.0

        if test < 0.0:
            phi = np.acos(np.sqrt(-27.0 * b * b / 4.0 / a / a / a))
            x1 = 2.0 * np.sqrt(-a / 3.0) * np.cos(phi / 3.0)
            x2 = 2.0 * np.sqrt(-a / 3.0) * np.cos(phi / 3.0 + np.pi * 2.0 / 3.0)
            x3 = 2.0 * np.sqrt(-a / 3.0) * np.cos(phi / 3.0 + np.pi * 4.0 / 3.0)
            if b > 0.0:
                x1 *= -1.0
                x2 *= -1.0
                x3 *= -1.0

        s1 = x1 - p / 3.0
        s2 = x2 - p / 3.0
        s3 = x3 - p / 3.0

        if s1 < s2 and s1 < s3:
            t1 = s2
            t2 = s3
        elif s2 < s1 and s2 < s3:
            t1 = s1
            t2 = s3
        else:
            t1 = s1
            t2 = s2

        b1 = np.asin(np.sqrt(t1))
        b2 = np.asin(np.sqrt(t2))

        betas = b1
        betaw = b2
        if b2 > b1:
            betas = b2
            betaw = b1

        if i == 0:
            return betaw
        if i == 1:
            return betas

    def mach1n_mach2n(self, M1, beta, gamma=1.4):
        M1n = M1 * np.sin(beta)
        M2n = np.sqrt(
            (1.0 + 0.5 * (gamma - 1.0) * M1n**2)
            / (gamma * M1n**2 - 0.5 * (gamma - 1.0))
        )
        return M1n, M2n

    def mach2(self, M1, theta, gamma=1.4):
        beta = self.beta(M1, theta)
        M1n, M2n = self.M1nM2n(M1, beta, gamma)
        M2 = M2n / np.sin(beta - theta)
        return M2

    def p2_p1(self, M1, theta, gamma=1.4):
        beta = self.beta(M1, theta)
        M1n = M1 * np.sin(beta)
        p2_p1 = 1.0 + 2.0 * gamma / (gamma + 1.0) * (M1n**2 - 1.0)
        return p2_p1



"""Provides class which connects formulae and data to UI elements through
Qt signals and slots.

"""

import sys

from qtpy import PYQT5, PYSIDE2, QtCore, QtGui, QtWidgets

from .. import formulae
from ..logger import logger

if PYQT5:
    from .base_pyqt import Ui_CalcDialog

    logger.debug("Using PyQt5 backend.")
elif PYSIDE2:
    from .base_pyside2 import Ui_CalcDialog

    logger.debug("Using PySide2 backend.")

Slot = QtCore.Slot


class CalcDialog(QtWidgets.QDialog):
    """
    Bridges all events in QApplication CalcApp to caeroc.formulae
    TODO: Error handling
    """

    def __init__(self, parent=None):
        """Initialize scalar values which keep track of different options in
        the GUI and setup the application user interface.

        """
        super().__init__(parent)

        self.ui = Ui_CalcDialog()
        self.ui.setupUi(self)
        # ----------Input------------
        self.key1 = self.ui.qcb1_input
        self.key2 = self.ui.qcb2_input
        self.input1 = 1.0
        self.input2 = 1.0
        self.gamma = 1.4
        self.autocalc = False
        self.mode = None
        self.on_qrb1_isen_pressed()
        # ---------Output------------
        self.table = self.ui.qtw_output
        self._setupModel()

        # -----Connect Signals-------
        self.autocalc = False
        if PYQT5:
            self.ui.qcb_autocalc.setChecked(self.autocalc)
            self.ui.qcb_autocalc.stateChanged.connect(self.on_qcb_autocalc_stateChanged)

    def _setupModel(self):
        """Sets up the model or the table where the output is displayed."""
        self.model = QtGui.QStandardItemModel(10, 2, self)
        self.model.setHeaderData(0, QtCore.Qt.Horizontal, "Parameter")
        self.model.setHeaderData(1, QtCore.Qt.Horizontal, "Value")

    def _set_keys_and_attrs(
        self, key1_to_attr, key2_to_attr, extra_attr_to_key, **kwargs
    ):
        """Reinitialize self.mode with an object of self.Mode - an alias for
        the formulae class. Also, clear the current options and add the related
        options for self.Mode.

        """

        if not isinstance(self.mode, self.Mode):
            self.key1.clear()
            self.key2.clear()
            self.key1_to_attr = key1_to_attr
            self.key2_to_attr = key2_to_attr
            self._set_attr_to_key(extra_attr_to_key)

            keys1 = self.key1_to_attr.keys()
            keys2 = self.key2_to_attr.keys()
            if not PYQT5:
                keys1 = list(keys1)
                keys2 = list(keys2)

            self.key1.addItems(keys1)
            self.key2.addItems(keys2)

    def _set_mode(self, **kwargs):
        if not isinstance(self.mode, self.Mode):
            self.mode = self.Mode(gamma=self.gamma, **kwargs)
            logger.debug(f"MODE: {self.mode.__doc__}")

    def _set_attr_to_key(self, extra=None):
        self.attr_to_key = {v: k for k, v in self.key1_to_attr.items()}
        self.attr_to_key.update({v: k for k, v in self.key2_to_attr.items()})
        self.attr_to_key.update(extra)

    @Slot()
    def on_qrb1_isen_pressed(self):
        """Run when radio button Isentropic is pressed."""
        self.Mode = formulae.isentropic.Isentropic
        key1_to_attr = {
            "M": "M",
            "p/p\u2080": "p_p0",
            "ρ/ρ\u2080": "rho_rho0",
            "T/T\u2080": "T_T0",
            "A/A*": "A_Astar",
        }
        key2_to_attr = {"-": None}
        extra_attr_to_key = {
            "Mt": "M*",
            "p_pt": "p/p*",
            "rho_rhot": "ρ/ρ*",
            "T_Tt": "T/T*",
        }
        self._set_keys_and_attrs(key1_to_attr, key2_to_attr, extra_attr_to_key)
        self._set_mode()

    @Slot()
    def on_qrb2_expa_pressed(self):
        """Run when radio button Expansion is pressed."""
        self.Mode = formulae.isentropic.Expansion
        key1_to_attr = {"θ (deg)": "theta_deg", "θ (rad)": "theta_rad"}
        key2_to_attr = {"M\u2081": "M_1", "ν\u2081 (rad)": "nu_1"}
        extra_attr_to_key = {
            "theta": "θ (rad)",
            "M_2": "M\u2082",
            "nu_2": "ν\u2082 (rad)",
            "p2_p1": "p\u2082/p\u2081",
            "rho2_rho1": "ρ\u2082/ρ\u2081",
            "T2_T1": "T\u2082/T\u2081",
        }
        self._set_keys_and_attrs(key1_to_attr, key2_to_attr, extra_attr_to_key)
        self._set_mode()

    @Slot()
    def on_qrb3_norm_pressed(self):
        """Run when radio button Normal Shock is pressed."""
        self.Mode = formulae.shock.NormalShock
        key1_to_attr = {
            "M\u2081": "M_1",
            "p\u2082/p\u2081": "p2_p1",
            "ρ\u2082/ρ\u2080": "rho2_rho1",
            "T\u2082/T\u2081": "T2_T1",
        }
        key2_to_attr = {"-": None}
        extra_attr_to_key = {
            "M_2": "M\u2082",
            "p02_p01": "p0\u2082/p0\u2081",
            "rho02_rho01": "ρ0\u2082/ρ0\u2080",
            "T02_T01": "T0\u2082/T0\u2081",
        }

        self._set_keys_and_attrs(key1_to_attr, key2_to_attr, extra_attr_to_key)
        kwargs = dict()
        key1 = self.key1.currentText()
        attr1 = key1_to_attr[key1]
        kwargs[attr1] = self.input1
        self._set_mode(**kwargs)

    @Slot()
    def on_qrb4_obli_pressed(self):
        """Run when radio button Oblique Shock is pressed."""
        logger.warn("MODE:Oblique Shock not implemented")

    @Slot()
    def on_qrb5_fann_pressed(self):
        """Run when radio button Fanno Flow is pressed."""
        logger.warn("MODE:Fanno Flow not implemented")

    @Slot()
    def on_qrb6_rayl_pressed(self):
        """Run when radio button Rayleigh Flow is pressed."""
        logger.warn("MODE:Rayleigh Flow not implemented")

    @Slot(float)
    def on_qdsb1_input_valueChanged(self, value):
        """Run when input value in the 1st input QDoubleSpinBox changes."""
        self.input1 = value
        if self.autocalc:
            self.on_qpb_calculate_released()

    @Slot(float)
    def on_qdsb2_input_valueChanged(self, value):
        """Run when input value in the QDoubleSpinBox changes."""
        self.input2 = value
        if self.autocalc:
            self.on_qpb_calculate_released()

    @Slot(float)
    def on_qdsb_gamma_valueChanged(self, value):
        """Run when input value in the QDoubleSpinBox changes."""
        self.gamma = value
        del self.mode
        self.mode = self.Mode(gamma=value)
        if self.autocalc:
            self.on_qpb_calculate_released()

    @Slot()
    def on_qcb_autocalc_stateChanged(self):
        """Run when the AutoCalculate QCheckBox state changes."""
        self.autocalc = self.ui.qcb_autocalc.isChecked()
        logger.debug(f"Auto calculate: {self.autocalc}")

    @Slot()
    def on_qpb_calculate_released(self):
        """Remove the rows and update the model with new data.
        Runs when the QPushButton is pressed and released.

        """
        attr1 = self.key1_to_attr[self.key1.currentText()]
        attr2 = self.key2_to_attr[self.key2.currentText()]
        if attr2 is None:
            kwargs = {attr1: self.input1}
        else:
            kwargs = {attr1: self.input1, attr2: self.input2}

        logger.info(kwargs)
        self.mode.calculate(**kwargs)
        logger.info(self.mode.data)

        # ------ Fill table --------------
        self.model.removeRows(
            0, self.model.rowCount(QtCore.QModelIndex()), QtCore.QModelIndex()
        )
        row = 0
        for attr in self.mode.keys:
            if self.mode.data[attr]:  # Not empty
                self.model.insertRows(row, 1, QtCore.QModelIndex())
                self.model.setData(
                    self.model.index(row, 0, QtCore.QModelIndex()),
                    self.attr_to_key[attr],
                )
                self.model.setData(
                    self.model.index(row, 1, QtCore.QModelIndex()),
                    str(self.mode.data[attr].pop()),
                )
                row += 1
        self.table.setModel(self.model)
        self.table.resizeColumnsToContents()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    calculator = CalcDialog()
    calculator.show()
    sys.exit(app.exec_())



from numbers import Number

import numpy as np

from ..logger import logger


class FormulaeBase:
    """Data management class for all flow quantities"""

    def __init__(self, *args, **kwargs):
        # self.keys = []
        self.data = {}
        self.minima = {}
        self.maxima = {}
        self._init_dataminmax()
        try:
            super().__init__(*args, **kwargs)
        except ValueError as e:
            logger.critical(e)

    def _init_dataminmax(self):
        """
        Initializes the data dictionary and the theoretical minima and maxima
        for flow quantities.
        """
        keys = self.keys
        if len(keys) == 0:
            raise ValueError("Unintialized list of keys")

        key_mach = ["M", "M_1", "M_2", "M_1n", "M_2n"]
        key_ratio = ["p_p0", "rho_rho0", "T_T0"]
        key_ang = ["nu", "theta"]
        inf = 1e10

        for k in keys:
            self.data[k] = []

            if k in key_mach:
                self.minima[k] = 0.0
                self.maxima[k] = 5.0
            elif k in key_ratio:
                self.minima[k] = 0.0
                self.maxima[k] = 1.0
            elif k in key_ang:
                self.minima[k] = 0.0
                if k == "theta":
                    self.maxima[k] = np.pi / 2
                else:
                    self.maxima[k] = (
                        (np.sqrt((self.gamma + 1.0) / (self.gamma - 1.0)) - 1)
                        * np.pi
                        / 2
                    )
            elif k == "A_At":
                self.minima[k] = 1.0
                self.maxima[k] = inf
            else:
                self.minima[k] = 0.0
                self.maxima[k] = inf

    def store(self, key, val):
        if key not in self.keys:
            raise ValueError(f"Unknown key: {key} not in {self.keys}")

        self._check_limits(key, val)
        self.data[key].append(val)

    def _check_limits(self, key, val):
        """Checks if the computed value is within the theoretical limits"""

        minimum = self.minima[key]
        maximum = self.maxima[key]
        if isinstance(val, np.ndarray) or isinstance(val, Number):
            if np.all(val < minimum) or np.all(val > maximum):
                logger.error(f"{key} = {val} must be between {minimum} and {maximum}")
        else:
            logger.critical(f"{val} is not a number.")



"""
Utilities for working with nozzles.
"""

import numpy as np

from caeroc._skaero.gasdynamics import isentropic, shocks


class Nozzle:
    """Class representing a nozzle."""

    def __init__(self, x, A):
        self.x = x
        self.A = A

    def plot(self, ax, where=None, interpolate=False, **kwargs):
        """Plots the nozzle geometry.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes where the nozzle geometry should be plotted.

        """
        radius = np.sqrt(self.A / np.pi)
        ax.set_xlim(self.x[0], self.x[-1])
        ax.fill_between(self.x, radius, -radius, where, interpolate, **kwargs)

    def solve_flow(self, fl, p_0, T_0, p_B):
        """Solves the flow through the nozzle.

        For the time being, it assumes isentropic flow and uses a quasi one
        dimensional model.

        Parameters
        ----------
        fl : IsentropicFlow
            Isentropic flow through the nozzle.
        p_0 : float
            Stagnation pressure at the reservoir.
        T_0 : float
            Stagnation temperature at the reservoir.
        p_B : float
            Back pressure at the exit section.

        Returns
        -------
        M : array
            Mach distribution along the nozzle.
        p : array
            Pressure distribution along the nozzle.

        Raises
        ------
        ValueError
            If the back pressure is higher than the reservoir pressure.

        Notes
        -----
        First of all, the function computes the isentropic conditions to
        discriminate if we have fully subsonic flow, chocked flow with shock
        inside nozzle or chocked fully supersonic flow.

        """
        p_ratio = p_B / p_0
        if p_ratio > 1.0:
            raise ValueError("Back pressure must be lower than reservoir pressure")

        # Exit and minimum area
        A_e = self.A[-1]
        A_min = np.min(self.A)
        i_min = np.argmin(self.A)

        # 1. Compute isentropic conditions at the exit
        # Points 1 (subsonic) and 2 (supersonic)
        M_e_1, M_e_2 = isentropic.mach_from_area_ratio(fl, A_e / A_min)

        pe_p0_1 = fl.p_p0(M_e_1)
        pe_p0_2 = fl.p_p0(M_e_2)

        # 2. Normal shock at the exit of the nozzle
        ns_e = shocks.NormalShock(M_e_2)
        pe_p0_3 = pe_p0_2 * ns_e.p2_p1

        # 3. Discriminate case
        M = np.empty_like(self.x)
        p = np.empty_like(self.x)

        if pe_p0_1 <= p_ratio <= 1.0:
            print("Fully subsonic flow")
            # 1. Compute M_e (inverse fl.p_p0(M))
            # 2. Compute area ratio from M_e (fl.A_Ac(M_e))
            # 3. The area ratio in each point is A_Ac = A * Ae_Ac / Ae
            #    so I can compute the Mach number distribution
        elif pe_p0_3 <= p_ratio < pe_p0_1:
            print("Shock inside the nozzle")
            # First I have to compute shock location and properties
        elif p_ratio < pe_p0_3:
            print("Fully isentropic subsonic-supersonic flow")
            # I already have Ac = A_min, so I can directly compute the Mach
            # number distribution
            Ac = A_min
            for i in range(i_min + 1):
                M[i] = isentropic.mach_from_area_ratio(fl, self.A[i] / Ac)[0]
            for i in range(i_min, len(M)):
                M[i] = isentropic.mach_from_area_ratio(fl, self.A[i] / Ac)[1]
            # Stagnation pressure does not change, as the flow is fully
            # isentropic
            p = p_0 * fl.p_p0(M)

        return M, p



"""
Decorators for formulae classes

"""

from functools import wraps


def storeresult(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        key = func.__name__
        result = func(self, *args, **kwargs)
        if "store" in kwargs.keys():
            if kwargs["store"] is True:
                self.store(key, result)
        else:
            for arg in args:
                if type(arg) is bool and arg is True:
                    self.store(key, result)

        return result

    return wrapper



"""
Shock waves.
"""

import inspect

import numpy as np
from scipy import optimize

from caeroc._skaero.gasdynamics.isentropic import IsentropicFlow, mach_angle


# Exceptions used in this module
class InvalidParametersError(Exception):
    pass


def max_deflection(M_1, gamma=1.4):
    """Returns maximum deflection angle and corresponding wave angle for given
    Mach number.

    Parameters
    ----------
    M_1 : float
        Upstream Mach number.
    gamma : float, optional
        Specific heat ratio, default 7 / 5.

    Returns
    -------
    theta : float
        Maximum deflection angle.
    beta : float
        Corresponding wave angle.

    """

    def eq(beta, M_1, gamma):
        os = _ShockClass(M_1, beta, gamma)
        return -os.theta

    mu = mach_angle(M_1)
    beta_theta_max = optimize.fminbound(eq, mu, np.pi / 2, args=(M_1, gamma), disp=0)
    os = _ShockClass(M_1, beta_theta_max, gamma)
    return os.theta, os.beta


def _ShockFactory(**kwargs):
    """Returns an object representing a shock wave.

    Parameters
    ----------
    gamma : float, optional
        Specific heat ratio, default 7 / 5.

    Examples
    --------
    >>> ss1 = Shock(M_1=1.5)  # Given upstream Mach number (default beta = 90°)
    >>> ss1.M_2
    0.70108874169309943
    >>> ss1.beta
    1.5707963267948966
    >>> ss1.theta
    0.0
    >>> ss2 = Shock(M_1=3.0, theta=np.radians(20.0), weak=True)
    >>> ss2.beta  # Notice it is an oblique shock
    0.6590997534071927

    TODO
    ----
    REMOVE MAGIC
    * This is a list of possible cases

      * M_1(, beta=np.pi / 2) -> _ShockClass(M_1, beta) (only direct case)
      * M_2(, beta=np.pi / 2)
      * p2_p1(, beta=np.pi / 2)
      * ...
      * M_1, theta(, weak=True)
      * M_2, theta(, weak=True)
      * p2_p1, theta(, weak=True)

    """
    kwargs.setdefault("gamma", 1.4)
    try:
        # We want a view of the keys, but the syntax changed in Python 3
        params = kwargs.viewkeys()
    except AttributeError:
        params = kwargs.keys()
    if "theta" not in params:
        kwargs.setdefault("beta", np.pi / 2)
        # ['X', 'beta', 'gamma']
        if len(params) != 3:
            raise InvalidParametersError("Invalid list of parameters")
    else:
        if "beta" in params:
            raise InvalidParametersError("Invalid list of parameters")
        kwargs.setdefault("weak", True)
        # ['X', 'theta', 'weak', 'gamma']
        if len(params) != 4:
            raise InvalidParametersError("Invalid list of parameters")

    # Here is the list of available resolution methods
    methods_list = [_from_deflection_angle]

    # And we generate a dictionary from it, indexed by their call arguments
    methods = {frozenset(inspect.getargspec(f)[0]): f for f in methods_list}
    # HACK, see http://stackoverflow.com/a/3999604/554319
    _k_class = frozenset(inspect.getargspec(_ShockClass.__init__)[0]) - set(["self"])
    methods[_k_class] = _ShockClass
    try:
        call_sig = frozenset(params)
        shock = methods[call_sig](**kwargs)
    except KeyError:
        raise NotImplementedError
    return shock


Shock = _ShockFactory


def _from_deflection_angle(M_1, theta, weak, gamma):
    """Returns oblique shock given upstream Mach number and deflection angle."""

    def eq(beta, M_1, theta, gamma):
        os = _ShockClass(M_1, beta, gamma)
        return os.theta - theta

    theta_max, beta_theta_max = max_deflection(M_1)
    if theta > theta_max:
        raise ValueError("No attached solution for this deflection angle")
    else:
        if weak:
            mu = mach_angle(M_1)
            beta = optimize.bisect(eq, mu, beta_theta_max, args=(M_1, theta, gamma))
        else:
            beta = optimize.bisect(
                eq, beta_theta_max, np.pi / 2, args=(M_1, theta, gamma)
            )

    return _ShockClass(M_1, beta, gamma)


class _ShockClass:
    """Class representing a shock."""

    def __init__(self, M_1, beta, gamma):
        mu = mach_angle(M_1)
        if beta < mu:
            raise ValueError(
                f"Shock wave angle must be higher than Mach angle {np.degrees(mu):.2f}°"
            )

        self.M_1 = M_1
        self.M_1n = M_1 * np.sin(beta) if beta != 0.0 else 0.0
        self.beta = beta
        self.gamma = gamma

    def __repr__(self):
        # FIXME: What if the object is returned from different parameters?
        return f"Shock(M_1={self.M_1!r}, beta={self.beta!r}, " f"gamma={self.gamma!r})"

    @property
    def theta(self):
        """Deflection angle of the shock."""
        if self.beta == mach_angle(self.M_1) or self.beta == np.pi / 2:
            theta = 0.0
        else:
            theta = np.arctan(
                2
                / np.tan(self.beta)
                * (np.sin(self.beta) ** 2 - 1 / self.M_1 / self.M_1)
                / (self.gamma + np.cos(2 * self.beta) + 2 / self.M_1 / self.M_1)
            )
        return theta

    @property
    def M_2n(self):
        """Normal Mach number behind the shock.

        FIXME: Raises ZeroDivisionError if M_1n == 0.0. Consistent?
        """
        M_2n = np.sqrt(
            (1 / (self.M_1n * self.M_1n) + (self.gamma - 1) / 2)
            / (self.gamma - (self.gamma - 1) / 2 / (self.M_1n * self.M_1n))
        )
        return M_2n

    @property
    def M_2(self):
        """Mach number behind the shock."""
        M_2 = self.M_2n / np.sin(self.beta - self.theta)
        return M_2

    @property
    def p2_p1(self):
        """Pressure ratio across the shock."""
        p2_p1 = 1 + 2 * self.gamma * (self.M_1n * self.M_1n - 1) / (self.gamma + 1)
        return p2_p1

    @property
    def rho2_rho1(self):
        """Density ratio accross the shock."""
        rho2_rho1 = (self.gamma + 1) / (2 / (self.M_1n * self.M_1n) + self.gamma - 1)
        return rho2_rho1

    @property
    def T2_T1(self):
        """Temperature ratio accross the shock."""
        T2_T1 = self.p2_p1 / self.rho2_rho1
        return T2_T1

    @property
    def p02_p01(self, fl=None):
        """Stagnation pressure ratio across the shock.

        Parameters
        ----------
        M_1 : float, optional
            Mach number behind the shock.
        fl : IsentropicFlow, optional
            To calculate stagnation conditions behind and after the shock.

        Returns
        -------
        p02_p01 : float
            Stagnation pressure ratio.

        """
        if fl is None:
            fl = IsentropicFlow(gamma=self.gamma)

        p02_p01 = fl.p_p0(self.M_1) / fl.p_p0(self.M_2) * self.p2_p1
        return p02_p01

    @property
    def rho02_rho01(self, fl=None):
        """Stagnation density ratio across the shock."""
        if fl is None:
            fl = IsentropicFlow(gamma=self.gamma)

        rho02_rho01 = fl.rho_rho0(self.M_1) / fl.rho_rho0(self.M_2) * self.rho2_rho1

        return rho02_rho01

    @property
    def T02_T01(self):
        """Stagnation temperature ratio across the shock."""
        T02_T01 = self.p02_p01 / self.rho02_rho01
        return T02_T01



import subprocess
import sys


def caeroc_app():
    import caeroc

    try:
        caeroc.launch()
    except AttributeError:
        app = caeroc.gui.CalcApp()
        app.run()


def caeroc_test():
    try:
        subprocess.call("caeroc-app", timeout=4)
    except subprocess.TimeoutExpired:
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == "__main__":
    caeroc_app()



"""
===========
scikit-aero
===========

Aeronautical engineering calculations in Python

"""

# Package version
__version__ = "0.2.dev0"

# Prellocate units module
import pint

units = pint.UnitRegistry()



import logging

try:
    import colorlog

    formatter = colorlog.ColoredFormatter(log_colors=colorlog.default_log_colors)
    handler = colorlog.StreamHandler()
    handler.setFormatter(formatter)
except ImportError:
    handler = logging.StreamHandler()
    print("Install colorlog!")


logger = logging.getLogger("caeroc-app")
logger.addHandler(handler)
logger.setLevel("DEBUG")
logger.info("caeroc: Compressible Aerodynamics Calculator")



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from caeroc.formulae.isentropic import Isentropic

isen = Isentropic()

M_max = 2.0
N = int(M_max * 100)
mach = list(np.linspace(0.1, M_max, N))
isen.data["M"] = mach

for i in range(N):
    isen.p_p0(mach[i])
    isen.rho_rho0(mach[i])
    isen.t_t0(mach[i])
    isen.a_at(mach[i])
    isen.p_pt(mach[i])
    isen.rho_rhot(mach[i])
    isen.t_tt(mach[i])

data = {}
for k in isen.keys:
    if isen.data[k]:  # Data is not empty
        data[k] = isen.data[k]

df = pd.DataFrame(data)
print(df)

# Plotting
plt.style.use("ggplot")
ax = df.plot(x="M")
fig = plt.gcf()
fig.show()



from caeroc.formulae.duct import Fanno, Rayleigh
from caeroc.formulae.isentropic import Expansion, Isentropic
from caeroc.formulae.shock import NormalShock, ObliqueShock
from caeroc.profiles.base import Airfoil



# Form implementation generated from reading ui file 'base.ui'
#
# Created by: PyQt5 UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_CalcDialog:
    def setupUi(self, CalcDialog):
        CalcDialog.setObjectName("CalcDialog")
        CalcDialog.resize(459, 692)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(CalcDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridWidget = QtWidgets.QWidget(CalcDialog)
        self.gridWidget.setObjectName("gridWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.gridWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.valueFrame = QtWidgets.QFrame(self.gridWidget)
        self.valueFrame.setObjectName("valueFrame")
        self.gridLayout = QtWidgets.QGridLayout(self.valueFrame)
        self.gridLayout.setObjectName("gridLayout")
        self.modeFrame = QtWidgets.QFrame(self.valueFrame)
        self.modeFrame.setObjectName("modeFrame")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.modeFrame)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.qrb4_obli = QtWidgets.QRadioButton(self.modeFrame)
        self.qrb4_obli.setObjectName("qrb4_obli")
        self.gridLayout_2.addWidget(self.qrb4_obli, 3, 2, 1, 1)
        self.qrb3_norm = QtWidgets.QRadioButton(self.modeFrame)
        self.qrb3_norm.setObjectName("qrb3_norm")
        self.gridLayout_2.addWidget(self.qrb3_norm, 2, 2, 1, 1)
        self.qrb1_isen = QtWidgets.QRadioButton(self.modeFrame)
        self.qrb1_isen.setChecked(True)
        self.qrb1_isen.setObjectName("qrb1_isen")
        self.gridLayout_2.addWidget(self.qrb1_isen, 2, 1, 1, 1)
        self.qrb5_fann = QtWidgets.QRadioButton(self.modeFrame)
        self.qrb5_fann.setObjectName("qrb5_fann")
        self.gridLayout_2.addWidget(self.qrb5_fann, 2, 3, 1, 1)
        self.qrb6_rayl = QtWidgets.QRadioButton(self.modeFrame)
        self.qrb6_rayl.setObjectName("qrb6_rayl")
        self.gridLayout_2.addWidget(self.qrb6_rayl, 3, 3, 1, 1)
        self.label = QtWidgets.QLabel(self.modeFrame)
        self.label.setEnabled(False)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 2, 0, 2, 1)
        self.qrb2_expa = QtWidgets.QRadioButton(self.modeFrame)
        font = QtGui.QFont()
        font.setStrikeOut(False)
        self.qrb2_expa.setFont(font)
        self.qrb2_expa.setObjectName("qrb2_expa")
        self.gridLayout_2.addWidget(self.qrb2_expa, 3, 1, 1, 1)
        self.gridLayout.addWidget(self.modeFrame, 0, 0, 1, 2)
        self.inputFrame = QtWidgets.QFrame(self.valueFrame)
        self.inputFrame.setObjectName("inputFrame")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.inputFrame)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.qcb1_input = QtWidgets.QComboBox(self.inputFrame)
        self.qcb1_input.setObjectName("qcb1_input")
        self.gridLayout_3.addWidget(self.qcb1_input, 5, 3, 1, 1)
        self.formFrame = QtWidgets.QFrame(self.inputFrame)
        self.formFrame.setObjectName("formFrame")
        self.formLayout = QtWidgets.QFormLayout(self.formFrame)
        self.formLayout.setObjectName("formLayout")
        self.qcb_autocalc = QtWidgets.QCheckBox(self.formFrame)
        self.qcb_autocalc.setObjectName("qcb_autocalc")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.qcb_autocalc)
        self.qpb_calculate = QtWidgets.QPushButton(self.formFrame)
        self.qpb_calculate.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.qpb_calculate.setAutoFillBackground(False)
        self.qpb_calculate.setObjectName("qpb_calculate")
        self.formLayout.setWidget(
            0, QtWidgets.QFormLayout.FieldRole, self.qpb_calculate
        )
        self.gridLayout_3.addWidget(self.formFrame, 7, 4, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.inputFrame)
        self.label_3.setEnabled(False)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setUnderline(False)
        font.setWeight(75)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.gridLayout_3.addWidget(self.label_3, 5, 0, 2, 1)
        self.qdsb2_input = QtWidgets.QDoubleSpinBox(self.inputFrame)
        self.qdsb2_input.setDecimals(8)
        self.qdsb2_input.setMinimum(-99.0)
        self.qdsb2_input.setSingleStep(0.1)
        self.qdsb2_input.setProperty("value", 1.0)
        self.qdsb2_input.setObjectName("qdsb2_input")
        self.gridLayout_3.addWidget(self.qdsb2_input, 6, 4, 1, 1)
        self.qcb2_input = QtWidgets.QComboBox(self.inputFrame)
        self.qcb2_input.setObjectName("qcb2_input")
        self.gridLayout_3.addWidget(self.qcb2_input, 6, 3, 1, 1)
        self.qdsb1_input = QtWidgets.QDoubleSpinBox(self.inputFrame)
        self.qdsb1_input.setDecimals(8)
        self.qdsb1_input.setMaximum(999.99)
        self.qdsb1_input.setSingleStep(0.1)
        self.qdsb1_input.setProperty("value", 1.0)
        self.qdsb1_input.setObjectName("qdsb1_input")
        self.gridLayout_3.addWidget(self.qdsb1_input, 5, 4, 1, 1)
        self.qdsb_gamma = QtWidgets.QDoubleSpinBox(self.inputFrame)
        self.qdsb_gamma.setSingleStep(0.1)
        self.qdsb_gamma.setProperty("value", 1.4)
        self.qdsb_gamma.setObjectName("qdsb_gamma")
        self.gridLayout_3.addWidget(self.qdsb_gamma, 7, 3, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.inputFrame)
        self.label_2.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.label_2.setTextFormat(QtCore.Qt.PlainText)
        self.label_2.setAlignment(
            QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing | QtCore.Qt.AlignVCenter
        )
        self.label_2.setObjectName("label_2")
        self.gridLayout_3.addWidget(self.label_2, 7, 0, 1, 1)
        self.gridLayout.addWidget(self.inputFrame, 2, 0, 4, 2)
        self.verticalLayout.addWidget(self.valueFrame)
        self.outputLayout = QtWidgets.QGridLayout()
        self.outputLayout.setObjectName("outputLayout")
        self.label_4 = QtWidgets.QLabel(self.gridWidget)
        self.label_4.setEnabled(False)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        font.setKerning(True)
        self.label_4.setFont(font)
        self.label_4.setAcceptDrops(False)
        self.label_4.setObjectName("label_4")
        self.outputLayout.addWidget(self.label_4, 0, 0, 1, 1)
        self.qtw_output = QtWidgets.QTableView(self.gridWidget)
        self.qtw_output.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.qtw_output.sizePolicy().hasHeightForWidth())
        self.qtw_output.setSizePolicy(sizePolicy)
        self.qtw_output.setAutoFillBackground(False)
        self.qtw_output.setFrameShape(QtWidgets.QFrame.Box)
        self.qtw_output.setFrameShadow(QtWidgets.QFrame.Plain)
        self.qtw_output.setObjectName("qtw_output")
        self.qtw_output.horizontalHeader().setCascadingSectionResizes(False)
        self.outputLayout.addWidget(self.qtw_output, 1, 0, 1, 1)
        self.verticalLayout.addLayout(self.outputLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(self.gridWidget)
        self.buttonBox.setStandardButtons(
            QtWidgets.QDialogButtonBox.Close | QtWidgets.QDialogButtonBox.Reset
        )
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)
        self.verticalLayout_2.addWidget(self.gridWidget)

        self.retranslateUi(CalcDialog)
        self.buttonBox.accepted.connect(CalcDialog.accept)  # type: ignore
        self.buttonBox.rejected.connect(CalcDialog.reject)  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(CalcDialog)

    def retranslateUi(self, CalcDialog):
        _translate = QtCore.QCoreApplication.translate
        CalcDialog.setWindowTitle(_translate("CalcDialog", "Caeroc Calculator"))
        CalcDialog.setToolTip(
            _translate("CalcDialog", "<html><head/><body><p>Input</p></body></html>")
        )
        self.qrb4_obli.setText(_translate("CalcDialog", "Ob&lique Shock"))
        self.qrb3_norm.setText(_translate("CalcDialog", "&Normal Shock"))
        self.qrb1_isen.setText(_translate("CalcDialog", "Isen&tropic"))
        self.qrb5_fann.setText(_translate("CalcDialog", "Fanno F&low"))
        self.qrb6_rayl.setText(_translate("CalcDialog", "Ra&yleigh Flow"))
        self.label.setToolTip(_translate("CalcDialog", "Choose a formula set"))
        self.label.setText(_translate("CalcDialog", "Mode"))
        self.qrb2_expa.setText(_translate("CalcDialog", "E&xpansion"))
        self.qcb_autocalc.setText(_translate("CalcDialog", "AutoCalculate"))
        self.qpb_calculate.setText(_translate("CalcDialog", "Calculate"))
        self.label_3.setText(_translate("CalcDialog", "Input"))
        self.label_2.setText(_translate("CalcDialog", "ɣ ="))
        self.label_4.setToolTip(_translate("CalcDialog", "Tabulated output"))
        self.label_4.setText(_translate("CalcDialog", "Result"))



################################################################################
## Form generated from reading UI file 'base.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CalcDialog:
    def setupUi(self, CalcDialog):
        if not CalcDialog.objectName():
            CalcDialog.setObjectName("CalcDialog")
        CalcDialog.resize(459, 692)
        self.verticalLayout_2 = QVBoxLayout(CalcDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridWidget = QWidget(CalcDialog)
        self.gridWidget.setObjectName("gridWidget")
        self.verticalLayout = QVBoxLayout(self.gridWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.valueFrame = QFrame(self.gridWidget)
        self.valueFrame.setObjectName("valueFrame")
        self.gridLayout = QGridLayout(self.valueFrame)
        self.gridLayout.setObjectName("gridLayout")
        self.modeFrame = QFrame(self.valueFrame)
        self.modeFrame.setObjectName("modeFrame")
        self.gridLayout_2 = QGridLayout(self.modeFrame)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.qrb4_obli = QRadioButton(self.modeFrame)
        self.qrb4_obli.setObjectName("qrb4_obli")

        self.gridLayout_2.addWidget(self.qrb4_obli, 3, 2, 1, 1)

        self.qrb3_norm = QRadioButton(self.modeFrame)
        self.qrb3_norm.setObjectName("qrb3_norm")

        self.gridLayout_2.addWidget(self.qrb3_norm, 2, 2, 1, 1)

        self.qrb1_isen = QRadioButton(self.modeFrame)
        self.qrb1_isen.setObjectName("qrb1_isen")
        self.qrb1_isen.setChecked(True)

        self.gridLayout_2.addWidget(self.qrb1_isen, 2, 1, 1, 1)

        self.qrb5_fann = QRadioButton(self.modeFrame)
        self.qrb5_fann.setObjectName("qrb5_fann")

        self.gridLayout_2.addWidget(self.qrb5_fann, 2, 3, 1, 1)

        self.qrb6_rayl = QRadioButton(self.modeFrame)
        self.qrb6_rayl.setObjectName("qrb6_rayl")

        self.gridLayout_2.addWidget(self.qrb6_rayl, 3, 3, 1, 1)

        self.label = QLabel(self.modeFrame)
        self.label.setObjectName("label")
        self.label.setEnabled(False)
        font = QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)

        self.gridLayout_2.addWidget(self.label, 2, 0, 2, 1)

        self.qrb2_expa = QRadioButton(self.modeFrame)
        self.qrb2_expa.setObjectName("qrb2_expa")
        font1 = QFont()
        font1.setStrikeOut(False)
        self.qrb2_expa.setFont(font1)

        self.gridLayout_2.addWidget(self.qrb2_expa, 3, 1, 1, 1)

        self.gridLayout.addWidget(self.modeFrame, 0, 0, 1, 2)

        self.inputFrame = QFrame(self.valueFrame)
        self.inputFrame.setObjectName("inputFrame")
        self.gridLayout_3 = QGridLayout(self.inputFrame)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.qcb1_input = QComboBox(self.inputFrame)
        self.qcb1_input.setObjectName("qcb1_input")

        self.gridLayout_3.addWidget(self.qcb1_input, 5, 3, 1, 1)

        self.formFrame = QFrame(self.inputFrame)
        self.formFrame.setObjectName("formFrame")
        self.formLayout = QFormLayout(self.formFrame)
        self.formLayout.setObjectName("formLayout")
        self.qcb_autocalc = QCheckBox(self.formFrame)
        self.qcb_autocalc.setObjectName("qcb_autocalc")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.qcb_autocalc)

        self.qpb_calculate = QPushButton(self.formFrame)
        self.qpb_calculate.setObjectName("qpb_calculate")
        self.qpb_calculate.setLayoutDirection(Qt.LeftToRight)
        self.qpb_calculate.setAutoFillBackground(False)

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.qpb_calculate)

        self.gridLayout_3.addWidget(self.formFrame, 7, 4, 1, 1)

        self.label_3 = QLabel(self.inputFrame)
        self.label_3.setObjectName("label_3")
        self.label_3.setEnabled(False)
        font2 = QFont()
        font2.setPointSize(11)
        font2.setBold(True)
        font2.setUnderline(False)
        font2.setWeight(75)
        self.label_3.setFont(font2)

        self.gridLayout_3.addWidget(self.label_3, 5, 0, 2, 1)

        self.qdsb2_input = QDoubleSpinBox(self.inputFrame)
        self.qdsb2_input.setObjectName("qdsb2_input")
        self.qdsb2_input.setDecimals(8)
        self.qdsb2_input.setMinimum(-99.000000000000000)
        self.qdsb2_input.setSingleStep(0.100000000000000)
        self.qdsb2_input.setValue(1.000000000000000)

        self.gridLayout_3.addWidget(self.qdsb2_input, 6, 4, 1, 1)

        self.qcb2_input = QComboBox(self.inputFrame)
        self.qcb2_input.setObjectName("qcb2_input")

        self.gridLayout_3.addWidget(self.qcb2_input, 6, 3, 1, 1)

        self.qdsb1_input = QDoubleSpinBox(self.inputFrame)
        self.qdsb1_input.setObjectName("qdsb1_input")
        self.qdsb1_input.setDecimals(8)
        self.qdsb1_input.setMaximum(999.990000000000009)
        self.qdsb1_input.setSingleStep(0.100000000000000)
        self.qdsb1_input.setValue(1.000000000000000)

        self.gridLayout_3.addWidget(self.qdsb1_input, 5, 4, 1, 1)

        self.qdsb_gamma = QDoubleSpinBox(self.inputFrame)
        self.qdsb_gamma.setObjectName("qdsb_gamma")
        self.qdsb_gamma.setSingleStep(0.100000000000000)
        self.qdsb_gamma.setValue(1.400000000000000)

        self.gridLayout_3.addWidget(self.qdsb_gamma, 7, 3, 1, 1)

        self.label_2 = QLabel(self.inputFrame)
        self.label_2.setObjectName("label_2")
        self.label_2.setLayoutDirection(Qt.LeftToRight)
        self.label_2.setTextFormat(Qt.PlainText)
        self.label_2.setAlignment(Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter)

        self.gridLayout_3.addWidget(self.label_2, 7, 0, 1, 1)

        self.gridLayout.addWidget(self.inputFrame, 2, 0, 4, 2)

        self.verticalLayout.addWidget(self.valueFrame)

        self.outputLayout = QGridLayout()
        self.outputLayout.setObjectName("outputLayout")
        self.label_4 = QLabel(self.gridWidget)
        self.label_4.setObjectName("label_4")
        self.label_4.setEnabled(False)
        font3 = QFont()
        font3.setPointSize(11)
        font3.setBold(True)
        font3.setWeight(75)
        font3.setKerning(True)
        self.label_4.setFont(font3)
        self.label_4.setAcceptDrops(False)

        self.outputLayout.addWidget(self.label_4, 0, 0, 1, 1)

        self.qtw_output = QTableView(self.gridWidget)
        self.qtw_output.setObjectName("qtw_output")
        self.qtw_output.setEnabled(True)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.qtw_output.sizePolicy().hasHeightForWidth())
        self.qtw_output.setSizePolicy(sizePolicy)
        self.qtw_output.setAutoFillBackground(False)
        self.qtw_output.setFrameShape(QFrame.Box)
        self.qtw_output.setFrameShadow(QFrame.Plain)
        self.qtw_output.horizontalHeader().setCascadingSectionResizes(False)

        self.outputLayout.addWidget(self.qtw_output, 1, 0, 1, 1)

        self.verticalLayout.addLayout(self.outputLayout)

        self.buttonBox = QDialogButtonBox(self.gridWidget)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setStandardButtons(
            QDialogButtonBox.Close | QDialogButtonBox.Reset
        )

        self.verticalLayout.addWidget(self.buttonBox)

        self.verticalLayout_2.addWidget(self.gridWidget)

        self.retranslateUi(CalcDialog)
        self.buttonBox.accepted.connect(CalcDialog.accept)
        self.buttonBox.rejected.connect(CalcDialog.reject)

        QMetaObject.connectSlotsByName(CalcDialog)

    # setupUi

    def retranslateUi(self, CalcDialog):
        CalcDialog.setWindowTitle(
            QCoreApplication.translate("CalcDialog", "Caeroc Calculator", None)
        )
        # if QT_CONFIG(tooltip)
        CalcDialog.setToolTip(
            QCoreApplication.translate(
                "CalcDialog", "<html><head/><body><p>Input</p></body></html>", None
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.qrb4_obli.setText(
            QCoreApplication.translate("CalcDialog", "Ob&lique Shock", None)
        )
        self.qrb3_norm.setText(
            QCoreApplication.translate("CalcDialog", "&Normal Shock", None)
        )
        self.qrb1_isen.setText(
            QCoreApplication.translate("CalcDialog", "Isen&tropic", None)
        )
        self.qrb5_fann.setText(
            QCoreApplication.translate("CalcDialog", "Fanno F&low", None)
        )
        self.qrb6_rayl.setText(
            QCoreApplication.translate("CalcDialog", "Ra&yleigh Flow", None)
        )
        # if QT_CONFIG(tooltip)
        self.label.setToolTip(
            QCoreApplication.translate("CalcDialog", "Choose a formula set", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.label.setText(QCoreApplication.translate("CalcDialog", "Mode", None))
        self.qrb2_expa.setText(
            QCoreApplication.translate("CalcDialog", "E&xpansion", None)
        )
        self.qcb_autocalc.setText(
            QCoreApplication.translate("CalcDialog", "AutoCalculate", None)
        )
        self.qpb_calculate.setText(
            QCoreApplication.translate("CalcDialog", "Calculate", None)
        )
        self.label_3.setText(QCoreApplication.translate("CalcDialog", "Input", None))
        self.label_2.setText(QCoreApplication.translate("CalcDialog", "\u0263 =", None))
        # if QT_CONFIG(tooltip)
        self.label_4.setToolTip(
            QCoreApplication.translate("CalcDialog", "Tabulated output", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.label_4.setText(QCoreApplication.translate("CalcDialog", "Result", None))

    # retranslateUi



"""
Utilities for atmospheric calculations.

"""

import numpy as np

# effective earth's radius
R_Earth = 6356.7660e3  # m


def geometric_to_geopotential(z):
    """Returns geopotential altitude from geometric altitude.

    Parameters
    ----------
    z : array_like
        Geometric altitude in meters.

    Returns
    -------
    h : array_like
        Geopotential altitude in meters.

    """

    h = np.asarray(z) * R_Earth / (np.asarray(z) + R_Earth)
    return h


def geopotential_to_geometric(h):
    """Returns geometric altitude from geopotential altitude.

    Parameters
    ----------
    h : array_like
        Geopotential altitude in meters.


    Returns
    -------
    z : array_like
        Geometric altitude in meters.

    Notes
    -----
    Based on eq. 19 of the `U.S. 1976 Standard Atmosphere`_.

    .. _`U.S. 1976 Standard Atmosphere`: http://ntrs.nasa.gov/search.jsp?R=1977\0009539

    """

    z = np.asarray(h) * R_Earth / (R_Earth - np.asarray(h))
    return z



""" Isentropic properties. """

import numpy as np
import scipy as sp
from scipy.optimize import bisect, newton

from caeroc._skaero.util.decorators import implicit


def mach_angle(M):
    r"""Returns Mach angle given supersonic Mach number.

    .. math::

        \mu = \arcsin{\left ( \frac{1}{M} \right )}

    Parameters
    ----------
    M : float
        Mach number.

    Returns
    -------
    mu : float
        Mach angle.

    Raises
    ------
    ValueError
        If given Mach number is subsonic.

    """
    try:
        with np.errstate(invalid="raise"):
            mu = np.arcsin(1 / M)
    except FloatingPointError:
        raise ValueError("Mach number must be supersonic")
    return mu


def mach_from_area_ratio(A_Astar, fl=None):
    """Computes the Mach number given an area ratio asuming isentropic flow.

    Uses the relation between Mach number and area ratio for isentropic flow,
    and returns both the subsonic and the supersonic solution.

    Parameters
    ----------
    A_Astar : float
        Cross sectional area.
    fl : IsentropicFlow, optional
        Isentropic flow object, default flow with gamma = 7 / 5.

    Returns
    -------
    out : tuple of floats
        Subsonic and supersonic Mach number solution of the equation.

    Raises
    ------
    ValueError
        If the area ratio is less than 1.0 (the critical area is always the
        minimum).

    """
    if not fl:
        fl = IsentropicFlow(gamma=1.4)
    eq = implicit(fl.A_Astar)
    if A_Astar < 1.0:
        raise ValueError("Area ratio must be greater than 1")
    elif A_Astar == 1.0:
        M_sub = M_sup = 1.0
    else:
        M_sub = bisect(eq, 0.0, 1.0, args=(A_Astar,))
        M_sup = newton(eq, 2.0, args=(A_Astar,))

    return M_sub, M_sup


def mach_from_nu(nu, in_radians=True, gamma=1.4):
    r"""Computes the Mach number given a Prandtl-Meyer angle, :math:`\nu`.

    Uses the relation between Mach number and Prandtl-Meyer angle for
    isentropic flow, to iteratively compute and return the Mach number.

    Parameters
    ----------
    nu : float
        Prandtl-Meyer angle, by default in radians.
    in_radians : bool, optional
        When set as False, converts nu from degrees to radians.
    gamma : float, optional
        Specific heat ratio.

    Returns
    -------
    M : float
        Mach number corresponding to :math:`\nu`.

    Raises
    ------
    ValueError
        If :math:`\nu` is 0 or negative or above the theoretical maxima based on
        :math:`\gamma`.

    """
    if not in_radians:
        nu = np.radians(nu)

    nu_max = np.pi / 2.0 * (np.sqrt((gamma + 1.0) / (gamma - 1.0)) - 1)
    if nu <= 0.0 or nu >= nu_max:
        raise ValueError(
            f"Prandtl-Meyer angle must be between (0, {nu_max:f}) radians."
        )

    eq = implicit(PrandtlMeyerExpansion.nu)
    M = newton(eq, 2.0, args=(nu,))

    return M


class IsentropicFlow:
    """Class representing an isentropic gas flow.

    Isentropic flow is characterized by:

    * Viscous and heat conductivity effects are negligible.
    * No chemical or radioactive heat production.

    """

    def __init__(self, gamma=1.4):
        """Constructor of IsentropicFlow.

        Parameters
        ----------
        gamma : float, optional
            Specific heat ratio, default 7 / 5.

        """
        self.gamma = gamma

    def p_p0(self, M):
        r"""Pressure ratio from Mach number.

        .. math::
            \left ( \frac{P}{P_{0}} \right ) = \left ( \frac{T}{T_{0}} \right )^{\frac{\gamma}{(\gamma - 1)}}

        Parameters
        ----------
        M : array_like
            Mach number.

        Returns
        -------
        p_p0 : array_like
            Pressure ratio.

        """

        M = np.asanyarray(M)
        p_p0 = self.T_T0(M) ** (self.gamma / (self.gamma - 1))

        return p_p0

    def rho_rho0(self, M):
        r"""Density ratio from Mach number.

        .. math::
            \left ( \frac{\rho}{\rho_{0}} \right ) = \left ( \frac{T}{T_{0}} \right )^{\frac{1}{(\gamma - 1)}}

        Parameters
        ----------
        M : array_like
            Mach number.

        Returns
        -------
        rho_rho0 : array_like
            Density ratio.

        """

        M = np.asanyarray(M)
        rho_rho0 = self.T_T0(M) ** (1 / (self.gamma - 1))
        return rho_rho0

    def T_T0(self, M):
        r"""Temperature ratio from Mach number.

        .. math::
            \left ( \frac{T}{T_{0}} \right ) = \left (1 + \frac{\gamma - 1}{2}M^{2} \right )^{-1}

        Parameters
        ----------
        M : array_like
            Mach number.

        Returns
        -------
        T_T0 : array_like
            Temperature ratio.

        """
        M = np.asanyarray(M)
        T_T0 = 1 / (1 + (self.gamma - 1) * M * M / 2)
        return T_T0

    def A_Astar(self, M):
        """Area ratio from Mach number.

        Duct area divided by critial area given Mach number.

        Parameters
        ----------
        M : array_like
            Mach number.

        Returns
        -------
        A_Astar : array_like
            Area ratio.

        """
        M = np.asanyarray(M)
        # If there is any zero entry, NumPy array division gives infinity,
        # which is correct.
        with np.errstate(divide="ignore"):
            A_Astar = (2 / self.T_T0(M) / (self.gamma + 1)) ** (
                (self.gamma + 1) / (2 * (self.gamma - 1))
            ) / M
        return A_Astar

    def a_a0(self, M):
        """Speed of sound ratio from Mach number.

        Parameters
        ----------
        M: array_like
            Mach number.

        Returns
        -------
        a_a0: array_like
            Speed of sound ratio.
        """

        M = np.asarray(M)
        a_a0 = self.T_T0(M) ** 0.5

        return a_a0


class PrandtlMeyerExpansion:
    """Class representing a Prandtl-Meyer expansion fan."""

    @staticmethod
    def nu(M, gamma=1.4):
        r"""Prandtl-Meyer angle for a given Mach number.

        The result is given by evaluating the Prandtl-Meyer function.

        .. math::
            \nu = \sqrt{\frac{\gamma + 1}{\gamma - 1}} \tan^{-1}\left [ \sqrt{\frac{\gamma - 1}{\gamma + 1}(M^{2} - 1)} \right ] - \tan^{-1}(\sqrt{M^{2} - 1})

        Parameters
        ----------
        M : float
            Mach number.
        gamma : float, optional
            Specific heat ratio, default 7 / 5.

        Returns
        -------
        nu : float
            Prandtl-Meyer angle, in radians.

        Raises
        ------
        ValueError
            If Mach number is subsonic.

        """
        try:
            with np.errstate(invalid="raise"):
                sgpgm = np.sqrt((gamma + 1) / (gamma - 1))
                nu = sgpgm * np.arctan(np.sqrt(M * M - 1) / sgpgm) - np.arctan(
                    np.sqrt(M * M - 1)
                )
        except FloatingPointError:
            raise ValueError("Mach number must be supersonic")
        return nu

    def __init__(self, M_1, theta, fl=None, gamma=1.4):
        """Constructor of PrandtlMeyerExpansion.

        Parameters
        ----------
        M_1 : float
            Upstream Mach number.
        theta : float
            Deflection angle, in radians.
        fl : IsentropicFlow, optional.
            Flow to be expanded
        gamma : float, optional
            Specific heat ratio, default value = 7 / 5.

        Raises
        ------
        ValueError
            If given Mach number is subsonic.

        """
        if not fl:
            fl = IsentropicFlow(gamma=gamma)
        nu_max = PrandtlMeyerExpansion.nu(np.inf, fl.gamma) - PrandtlMeyerExpansion.nu(
            M_1, fl.gamma
        )
        if theta > nu_max:
            raise ValueError(
                f"Deflection angle must be lower than maximum {np.degrees(nu_max):.2f}°"
            )
        self.M_1 = M_1
        self.theta = theta
        self.fl = fl

    @property
    def nu_1(self):
        """Upstream Prandtl-Meyer angle."""
        return PrandtlMeyerExpansion.nu(self.M_1, self.fl.gamma)

    @property
    def nu_2(self):
        """Downstream Prandtl-Meyer angle."""
        return self.nu_1 + self.theta

    @property
    def M_2(self):
        """Downstream Mach number."""
        return mach_from_nu(nu=self.nu_2, gamma=self.fl.gamma)

    @property
    def mu_1(self):
        """Angle of forward Mach line."""
        return mach_angle(self.M_1)

    @property
    def mu_2(self):
        """Angle of rearward Mach line."""
        return mach_angle(self.M_2)

    @property
    def p2_p1(self):
        """Pressure ratio across the expansion fan."""
        p2_p1 = self.fl.p_p0(self.M_2) / self.fl.p_p0(self.M_1)
        return p2_p1

    @property
    def T2_T1(self):
        """Temperature ratio across the expansion fan."""
        T2_T1 = self.fl.T_T0(self.M_2) / self.fl.T_T0(self.M_1)
        return T2_T1

    @property
    def rho2_rho1(self):
        """Density ratio across the expansion fan."""
        return self.p2_p1 / self.T2_T1



"""
.. _U.S. 1976 Standard Atmosphere: http://ntrs.nasa.gov/search.jsp?R=19770009539

COESA model, based on the `U.S. 1976 Standard Atmosphere`_.

"""

import numpy as np
from scipy import constants, interpolate

from caeroc._skaero.atmosphere import util

# Constants and values : the following parameters are extracted from Notes
# reference (mainly Chap. 1.2.). Naming is consistent. WARNING : Some of these
# values are not exactly consistent with the 2010 CODATA Recommended Values of
# the Fundamental Physical constants that you can find for example in the
# scipy.constants module

# gas constant
Rs = 8.31432  # N m / (mol K), WARNING : different from the the 2010 CODATA
# Recommended Values of the Fundamental Physical Constants

# set of geopotential heights from table 2 of Notes reference
H = np.array([0.0, 11.0, 20.0, 32.0, 47.0, 51.0, 71.00, 84.85205]) * 1e3  # m

# set of molecular-scale temperature gradients from table 2 of Notes reference
LM = np.array([-6.5, 0.0, 1.0, 2.8, 0.0, -2.8, -2.0, 0.0]) * 1e-3  # K / m

f_LM = interpolate.interp1d(H, LM, kind="zero")

# K, standard sea-level temperature
T_0 = 288.15  # K

# mean molecular-weight at sea-level
M_0 = 28.9644e-3  # kg / mol

# set of geopotential heights from table 8 of Notes reference
H2 = np.array(
    [
        0.0,
        79005.7,
        79493.3,
        79980.8,
        80468.2,
        80955.7,
        81443.0,
        81930.2,
        82417.3,
        82904.4,
        83391.4,
        83878.4,
        84365.2,
        84852.05,
    ]
)  # m

# set of molecular weight ratios from table 8 of Notes reference
M_o_M0 = np.array(
    [
        1.0,
        1.0,
        0.999996,
        0.999989,
        0.999971,
        0.999941,
        0.999909,
        0.999870,
        0.999829,
        0.999786,
        0.999741,
        0.999694,
        0.999641,
        0.999579,
    ]
)  # -

f_M_o_M0 = interpolate.interp1d(H2, M_o_M0)

# set of pressures and temperatures (initialization)
P = np.array([constants.atm])  # Pa
TM = np.array([T_0])  # K

for k in range(1, len(H)):
    # from eq. [23] of Notes reference
    TM = np.append(TM, TM[-1] + f_LM(H[k - 1]) * (H[k] - H[k - 1]))
    if f_LM(H[k - 1]) == 0.0:
        # from eq. [33b] of Notes reference
        P = np.append(
            P, P[-1] * np.exp(-constants.g * M_0 * (H[k] - H[k - 1]) / (Rs * TM[-2]))
        )
    else:
        # from eq. [33a] of Notes reference
        P = np.append(
            P,
            P[-1] * (TM[-2] / (TM[-1])) ** (constants.g * M_0 / (Rs * f_LM(H[k - 1]))),
        )

f_TM = interpolate.interp1d(H, TM, kind="zero")
f_P = interpolate.interp1d(H, P, kind="zero")
f_H = interpolate.interp1d(H, H, kind="zero")


def table(x, kind="geopotential"):
    """Computes table of COESA atmosphere properties.

    Returns temperature, pressure and density COESA values at the given
    altitude.

    Parameters
    ----------
    x : array_like
       Geopotential or geometric altitude (depending on kind) given in meters.
    kind : str
       Specifies the kind of interpolation as altitude x ('geopotential' or 'geometric'). Default is 'geopotential'

    Returns
    -------
    h : array_like
        Given geopotential altitude in meters.
    T : array_like
        Temperature in Kelvin.
    p : array_like
        Pressure in Pascal.
    rho : array_like
        Density in kilograms per cubic meter.

    Note
    ----
    Based on the U.S. 1976 Standard Atmosphere.

    """

    # check the kind of altitude and raise an exception if necessary
    if kind == "geopotential":
        alt = x
    elif kind == "geometric":
        alt = util.geometric_to_geopotential(x)
    else:
        raise ValueError(
            f"{kind} is unsupported: Use either geopotential or " "geometric."
        )

    h = np.asarray(alt)

    # check if altitude is out of bound and raise an exception if necessary
    if (h < H[0]).any() or (h > H[-1]).any():
        raise ValueError(
            "the given altitude x is out of bound, this module is "
            "currently only valid for a geometric altitude between 0. and 86000. m"
        )

    # K, molecule-scale temperature from eq. [23] of Notes reference
    tm = f_TM(h) + f_LM(h) * (h - f_H(h))

    # K, absolute temperature from eq. [22] of Notes reference
    T = tm * f_M_o_M0(h)

    if h.shape:  # if h is not a 0-d array (like a scalar)
        # Pa, intialization of the pressure vector
        p = np.zeros(len(h))

        # points of h for which the molecular-scale temperature gradient is
        # zero
        zero_gradient = f_LM(h) == 0.0

        # points of h for which the molecular-scale temperature gradient is not
        # zero
        not_zero_gradient = f_LM(h) != 0.0

        # Pa, pressure from eq. [33b] of Notes reference
        p[zero_gradient] = f_P(h[zero_gradient]) * np.exp(
            -constants.g
            * M_0
            * (h[zero_gradient] - f_H(h[zero_gradient]))
            / (Rs * f_TM(h[zero_gradient]))
        )

        # Pa, pressure from eq. [33a] of Notes reference
        p[not_zero_gradient] = f_P(h[not_zero_gradient]) * (
            f_TM(h[not_zero_gradient])
            / (
                f_TM(h[not_zero_gradient])
                + f_LM(h[not_zero_gradient])
                * (h[not_zero_gradient] - f_H(h[not_zero_gradient]))
            )
        ) ** (constants.g * M_0 / (Rs * f_LM(h[not_zero_gradient])))

    else:
        if f_LM(h) == 0:
            # Pa, pressure from eq. [33b] of Notes reference
            p = f_P(h) * np.exp(-constants.g * M_0 * (h - f_H(h)) / (Rs * f_TM(h)))
        else:
            # Pa, pressure from eq. [33a] of Notes reference
            p = f_P(h) * (f_TM(h) / (f_TM(h) + f_LM(h) * (h - f_H(h)))) ** (
                constants.g * M_0 / (Rs * f_LM(h))
            )

    rho = p * M_0 / (Rs * tm)  # kg / m^3, mass density

    return alt, T, p, rho


def temperature(x, kind="geopotential"):
    """Computes air temperature for a given altitude using the U.S. standard atmosphere model

    Parameters
    ----------
    x : array_like
       Geopotential or geometric altitude (depending on kind) given in meters.
    kind : str
       Specifies the kind of interpolation as altitude x ('geopotential' or 'geometric'). Default is 'geopotential'

    Returns
    -------
    T : array_like
        Temperature in Kelvin.

    Note
    ----
    Based on the U.S. 1976 Standard Atmosphere.

    """

    T = table(x, kind)[1]
    return T


def pressure(x, kind="geopotential"):
    """Computes absolute pressure for a given altitude using the U.S. standard atmosphere model.

    Parameters
    ----------
    x : array_like
       Geopotential or geometric altitude (depending on kind) given in meters.
    kind : str
       Specifies the kind of interpolation as altitude x ('geopotential' or 'geometric'). Default is 'geopotential'

    Returns
    -------
    P : array_like
        Pressure in Pascal.

    Note
    ----
    Based on the U.S. 1976 Standard Atmosphere.

    """

    p = table(x, kind)[2]
    return p


def density(x, kind="geopotential"):
    """Computes air mass density for a given altitude using the U.S. standar atmosphere model.

    Parameters
    ----------
    x : array_like
       Geopotential or geometric altitude (depending on kind) given in meters.
    kind : str
       Specifies the kind of interpolation as altitude x ('geopotential' or 'geometric'). Default is 'geopotential'

    Returns
    -------
    rho : array_like
        Density in kilograms per cubic meter.

    Note
    ----
    Based on the U.S. 1976 Standard Atmosphere.

    """

    rho = table(x, kind)[3]
    return rho



"""
Created on Wed May 20 16:52:11 2015

@author: ashwin
"""

import numpy as np
import pylab as pl


class ProfilesBase:
    def __init__(self, N, length, bodytype=None):
        self.N = N
        self.L = length
        if bodytype is "airfoil":
            self.X, self.Y, self.angle = self.airfoil()
        else:
            self.X, self.Y, self.angle = self.circle()
        pl.plot(self.X, self.Y)
        pl.axis([-length, length, -length, length])
        self._non_dimensionalize(length)

    def _non_dimensionalize(self, scale):
        """Non dimensionalize all geometric parameters with the scale"""
        self.L = self.L / scale
        self.X = self.X / scale
        self.Y = self.Y / scale


class Nozzle(ProfilesBase):
    def nozzle(self):
        """Converging/Diverging Nozzle"""
        pass

    def lavalnozzle(self):
        """Conv. and Div. Nozzle"""
        pass


class Airfoil(ProfilesBase):
    def airfoil(self):
        """Biconvex airfoil"""
        L = self.L
        N = self.N
        R = 210
        theta_arc = np.arcsin(L / 2 / R) * 2
        pi = np.pi
        theta_up = np.linspace((pi + theta_arc) / 2, +(pi - theta_arc) / 2, N)
        # theta_up = theta[::-1]
        theta_down = np.linspace((3 * pi - theta_arc) / 2, +(3 * pi + theta_arc) / 2, N)

        X_up = R * np.cos(theta_up)
        Y_up = R * np.sin(theta_up)
        X_down = R * np.cos(theta_down)
        Y_down = R * np.sin(theta_down)

        shift_r = X_up[0]
        X_up -= shift_r
        X_down -= shift_r
        shift_up = Y_up[0]
        shift_down = Y_down[0]
        Y_up = Y_up - shift_up
        Y_down = Y_down - shift_down

        X = np.concatenate((X_up, X_down))
        Y = np.concatenate((Y_up, Y_down))
        slope_up = np.gradient(Y_up, 1) / np.gradient(X_up, 1)
        slope_down = np.gradient(Y_down, 1) / np.gradient(X_down, 1)
        angle = np.arctan(np.concatenate((slope_up, slope_down)))
        return X, Y, angle

    def circle(self):
        theta = np.linspace(0, 2 * np.pi, self.N)
        R = self.L / 2
        X = R * np.cos(theta)
        Y = R * np.sin(theta)
        angle = theta - np.pi / 2
        return X, Y, angle


class Duct(ProfilesBase):
    def duct(self):
        """Constant area duct/pipe for Fanno & Rayleigh flows"""
        pass
